@@ -36,6 +36,17 @@
 #include <gdk-pixbuf/gdk-pixbuf.h>
 
 
+/**
+ * After how long does our long-poller time out?
+ */
+#define LP_TIMEOUT GNUNET_TIME_relative_multiply (GNUNET_TIME_UNIT_MINUTES, 5)
+
+/**
+ * Next time we schedule the #long_task.
+ */
+static struct GNUNET_TIME_Absolute long_next;
+
+
 /**
  * Are we currently processing an action?
  */
@@ -100,6 +111,7 @@ action_continent_selecting (void)
                                NULL, NULL))
         {
           GNUNET_break (0);
+          GNUNET_JSON_parse_free (spec);
           continue;
         }
 
@@ -135,7 +147,6 @@ action_continent_selecting (void)
   AG_show ("anastasis_gtk_main_window_forward_button");
   AG_show ("anastasis_gtk_main_control_vbox");
   AG_show ("anastasis_gtk_continent_frame");
-  AG_show ("anastasis_gtk_continent_selection_image");
 }
 
 
@@ -181,6 +192,7 @@ action_country_selecting (void)
                              NULL, NULL))
       {
         GNUNET_break (0);
+        GNUNET_JSON_parse_free (spec);
         continue;
       }
 
@@ -229,10 +241,7 @@ action_country_selecting (void)
   }
   AG_show ("anastasis_gtk_main_window_prev_button");
   AG_show ("anastasis_gtk_main_window_forward_button");
-  AG_show ("anastasis_gtk_country_selection_image");
   AG_show ("anastasis_gtk_continent_frame");
-  AG_show ("anastasis_gtk_continent_selection_image");
-  AG_show ("anastasis_gtk_country_selection_image");
 }
 
 
@@ -613,6 +622,7 @@ action_user_attributes_collecting (void)
                                     attr_type,
                                     val);
       }
+      GNUNET_JSON_parse_free (spec);
     }
   }
 
@@ -636,7 +646,6 @@ action_user_attributes_collecting (void)
   AG_show ("anastasis_gtk_main_window_forward_button");
   AG_show ("anastasis_gtk_identity_frame");
   AG_focus ("anastasis_gtk_ia_full_name_entry");
-  AG_show ("anastasis_gtk_user_attributes_image");
 }
 
 
@@ -799,7 +808,6 @@ action_authentications_editing (void)
   AG_show ("anastasis_gtk_main_window_prev_button");
   AG_show ("anastasis_gtk_main_window_forward_button");
   AG_show ("anastasis_gtk_b_authentication_frame");
-  AG_show ("anastasis_gtk_b_authentication_methods_image");
 }
 
 
@@ -1040,7 +1048,6 @@ action_policies_reviewing (void)
   AG_show ("anastasis_gtk_backup_progress_scrolled_window");
   AG_hide ("anastasis_gtk_recovery_progress_scrolled_window");
   AG_show ("anastasis_gtk_b_policy_frame");
-  AG_show ("anastasis_gtk_b_policies_image");
   AG_show ("anastasis_gtk_main_control_vbox");
   AG_show ("anastasis_gtk_main_window_prev_button");
   AG_show ("anastasis_gtk_main_window_forward_button");
@@ -1321,7 +1328,6 @@ action_secret_editing (void)
   AG_show ("anastasis_gtk_backup_progress_scrolled_window");
   AG_hide ("anastasis_gtk_recovery_progress_scrolled_window");
   AG_show ("anastasis_gtk_enter_secret_frame");
-  AG_show ("anastasis_gtk_enter_secret_image");
   AG_show ("anastasis_gtk_main_control_vbox");
   AG_show ("anastasis_gtk_main_window_prev_button");
   AG_show ("anastasis_gtk_main_window_forward_button");
@@ -1527,7 +1533,6 @@ action_truths_paying (void)
     json_decref (args);
   }
   AG_show ("anastasis_gtk_pay_frame");
-  AG_show ("anastasis_gtk_pay_image");
   AG_show ("anastasis_gtk_main_control_vbox");
   AG_show ("anastasis_gtk_progress_vbox");
   AG_progress_update ();
@@ -1610,7 +1615,6 @@ action_policies_paying (void)
     json_decref (args);
   }
   AG_show ("anastasis_gtk_pay_frame");
-  AG_show ("anastasis_gtk_pay_image");
   AG_show ("anastasis_gtk_main_control_vbox");
   AG_show ("anastasis_gtk_progress_vbox");
   AG_progress_update ();
@@ -1705,7 +1709,6 @@ action_backup_finished (void)
   AG_hide ("anastasis_gtk_success_recovery_box");
   AG_show ("anastasis_gtk_success_backup_label");
   AG_hide ("anastasis_gtk_success_recovery_box");
-  AG_show ("anastasis_gtk_completed_image");
   AG_show ("anastasis_gtk_main_control_vbox");
   AG_hide ("anastasis_gtk_main_window_save_as_button");
   AG_show ("anastasis_gtk_restart_button");
@@ -1840,6 +1843,8 @@ translate_state (const char *state)
       .out = _ ("wait, tries exceeded") },
     { .in = "authentication-timeout",
       .out = _ ("awaiting completion of authentication process") },
+    { .in = "external-instructions",
+      .out = _ ("challenge-specific action required") },
     { .in = NULL,
       .out = NULL }
   };
@@ -2195,7 +2200,92 @@ action_secret_selecting (void)
 
 
 /**
- * FIXME.
+ * Function called with the results of #ANASTASIS_redux_action on "poll".
+ *
+ * @param cls closure, NULL
+ * @param error_code Error code
+ * @param response new state as result or config information of a provider
+ */
+static void
+long_action_cb (void *cls,
+                enum TALER_ErrorCode error_code,
+                json_t *response);
+
+
+/**
+ * Schedules the 'poll' action.
+ *
+ * @param cls NULL
+ */
+static void
+long_task (void *cls)
+{
+  json_t *tspec;
+
+  (void) cls;
+  AG_long_task = NULL;
+  if (GNUNET_TIME_absolute_is_future (long_next))
+  {
+    AG_long_task = GNUNET_SCHEDULER_add_at (long_next,
+                                            &long_task,
+                                            NULL);
+    return;
+  }
+  long_next = GNUNET_TIME_relative_to_absolute (LP_TIMEOUT);
+  tspec = GNUNET_JSON_PACK (
+    GNUNET_JSON_pack_time_rel ("timeout",
+                               LP_TIMEOUT));
+  AG_long_action
+    = ANASTASIS_redux_action (AG_redux_state,
+                              "poll",
+                              tspec,
+                              &long_action_cb,
+                              NULL);
+  json_decref (tspec);
+}
+
+
+/**
+ * Function called with the results of #ANASTASIS_redux_action on "poll".
+ *
+ * @param cls closure, NULL
+ * @param error_code Error code
+ * @param response new state as result or config information of a provider
+ */
+static void
+long_action_cb (void *cls,
+                enum TALER_ErrorCode error_code,
+                json_t *response)
+{
+  AG_long_action = NULL;
+  switch (error_code)
+  {
+  case TALER_EC_NONE:
+    /* continued below */
+    break;
+  default:
+    GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
+                "'poll' action failed: %s\n",
+                TALER_ErrorCode_get_hint (error_code));
+    /* simply try again */
+    AG_long_task = GNUNET_SCHEDULER_add_now (&long_task,
+                                             NULL);
+    return;
+  }
+  if (NULL != AG_ra)
+  {
+    GNUNET_break (0);
+    ANASTASIS_redux_action_cancel (AG_ra);
+  }
+  AG_action_cb (NULL,
+                TALER_EC_NONE,
+                response);
+}
+
+
+/**
+ * The user must select the next challenge to solve
+ * during the recovery process.
  */
 static void
 action_challenge_selecting (void)
@@ -2223,6 +2313,7 @@ action_challenge_selecting (void)
       const char *type;
       const char *uuid;
       struct TALER_Amount cost;
+      bool async = false;
       struct GNUNET_JSON_Specification spec[] = {
         GNUNET_JSON_spec_string ("instructions",
                                  &instructions),
@@ -2232,6 +2323,9 @@ action_challenge_selecting (void)
                                  &provider),
         GNUNET_JSON_spec_string ("uuid",
                                  &uuid),
+        GNUNET_JSON_spec_mark_optional (
+          GNUNET_JSON_spec_bool ("async",
+                                 &async)),
         GNUNET_JSON_spec_end ()
       };
 
@@ -2264,7 +2358,14 @@ action_challenge_selecting (void)
                                  uuid,
                                  NULL))
         continue;
-
+      if (async &&
+          (NULL == AG_long_task) )
+      {
+        long_next = GNUNET_TIME_UNIT_ZERO_ABS;
+        AG_long_task
+          = GNUNET_SCHEDULER_add_now (&long_task,
+                                      NULL);
+      }
       gtk_list_store_insert_with_values (
         ls,
         NULL,
@@ -2434,6 +2535,10 @@ action_challenge_selecting (void)
 }
 
 
+/**
+ * An Anastasis provider requires payment for a challenge.
+ * Give opportunity to the user to pay.
+ */
 static void
 action_challenge_paying (void)
 {
@@ -2534,7 +2639,6 @@ action_challenge_paying (void)
   AG_show ("anastasis_gtk_recovery_progress_scrolled_window");
   AG_hide ("anastasis_gtk_backup_progress_scrolled_window");
   AG_show ("anastasis_gtk_pay_frame");
-  AG_show ("anastasis_gtk_pay_image");
   AG_show ("anastasis_gtk_main_control_vbox");
   AG_sensitive ("anastasis_gtk_main_window_prev_button");
   AG_show ("anastasis_gtk_main_window_prev_button");
@@ -2748,6 +2852,7 @@ diag_question (const json_t *details)
 
     uuid_str = json_string_value (json_object_get (details,
                                                    "uuid"));
+    /* Why do we do this? */
     render_feedback (builder,
                      "anastasis_gtk_c_question_error_label",
                      uuid_str);
@@ -2766,6 +2871,27 @@ static GtkDialog *
 diag_code (const json_t *details)
 {
   GtkBuilder *builder;
+  const char *instructions;
+  const char *uuid_str;
+  struct GNUNET_JSON_Specification spec[] = {
+    GNUNET_JSON_spec_string ("instructions",
+                             &instructions),
+    GNUNET_JSON_spec_string ("uuid",
+                             &uuid_str),
+    GNUNET_JSON_spec_end ()
+  };
+
+  if (GNUNET_OK !=
+      GNUNET_JSON_parse (details,
+                         spec,
+                         NULL, NULL))
+  {
+    GNUNET_break (0);
+    json_dumpf (details,
+                stderr,
+                JSON_INDENT (2));
+    return NULL;
+  }
 
   builder = GNUNET_GTK_get_new_builder ("anastasis_gtk_challenge_code.glade",
                                         NULL);
@@ -2776,25 +2902,16 @@ diag_code (const json_t *details)
   }
   {
     GtkLabel *label;
-    const char *instructions;
 
     label = GTK_LABEL (gtk_builder_get_object (builder,
                                                "challenge_instructions_label"));
-    instructions = json_string_value (json_object_get (details,
-                                                       "instructions"));
     gtk_label_set_text (label,
                         instructions);
-  }
-  {
-    GtkLabel *clabel;
-    const char *uuid_str;
-
-    clabel = GTK_LABEL (gtk_builder_get_object (builder,
-                                                "anastasis_gtk_c_challenge_label"));
-    uuid_str = json_string_value (json_object_get (details,
-                                                   "uuid"));
-    gtk_label_set_text (clabel,
+    label = GTK_LABEL (gtk_builder_get_object (builder,
+                                               "anastasis_gtk_c_challenge_label"));
+    gtk_label_set_text (label,
                         uuid_str);
+    /* Why do we do this? */
     render_feedback (builder,
                      "anastasis_gtk_c_code_error_label",
                      uuid_str);
@@ -2809,6 +2926,116 @@ diag_code (const json_t *details)
 }
 
 
+/**
+ * Create a dialog for the user to make an IBAN transfer.
+ *
+ * @param details details about the dialog to render
+ * @return dialog object
+ */
+static GtkDialog *
+diag_iban (const json_t *details)
+{
+  GtkBuilder *builder;
+  struct TALER_Amount amount;
+  const char *credit_iban;
+  const char *business;
+  const char *subject;
+  const char *uuid_str;
+  const char *debit_iban_hint;
+  struct GNUNET_JSON_Specification spec[] = {
+    GNUNET_JSON_spec_string ("uuid",
+                             &uuid_str),
+    GNUNET_JSON_spec_string ("instructions",
+                             &debit_iban_hint),
+    GNUNET_JSON_spec_end ()
+  };
+
+  if (GNUNET_OK !=
+      GNUNET_JSON_parse (details,
+                         spec,
+                         NULL, NULL))
+  {
+    GNUNET_break (0);
+    json_dumpf (details,
+                stderr,
+                JSON_INDENT (2));
+    return NULL;
+  }
+  {
+    json_t *cf = json_object_get (AG_redux_state,
+                                  "challenge_feedback");
+    json_t *ci = json_object_get (cf,
+                                  uuid_str);
+    json_t *cd = json_object_get (ci,
+                                  "details");
+    struct GNUNET_JSON_Specification ispec[] = {
+      TALER_JSON_spec_amount_any ("challenge_amount",
+                                  &amount),
+      GNUNET_JSON_spec_string ("credit_iban",
+                               &credit_iban),
+      GNUNET_JSON_spec_string ("business_name",
+                               &business),
+      GNUNET_JSON_spec_string ("wire_transfer_subject",
+                               &subject),
+      GNUNET_JSON_spec_end ()
+    };
+
+    if ( (NULL == cd) ||
+         (GNUNET_OK !=
+          GNUNET_JSON_parse (cd,
+                             ispec,
+                             NULL, NULL)) )
+    {
+      GNUNET_break (0);
+      json_dumpf (AG_redux_state,
+                  stderr,
+                  JSON_INDENT (2));
+      return NULL;
+    }
+  }
+
+  builder = GNUNET_GTK_get_new_builder ("anastasis_gtk_challenge_iban.glade",
+                                        NULL);
+  if (NULL == builder)
+  {
+    GNUNET_break (0);
+    return NULL;
+  }
+  {
+    GtkLabel *label;
+
+    label = GTK_LABEL (gtk_builder_get_object (builder,
+                                               "debit_account_label"));
+    gtk_label_set_text (label,
+                        debit_iban_hint);
+    label = GTK_LABEL (gtk_builder_get_object (builder,
+                                               "credit_account_label"));
+    gtk_label_set_text (label,
+                        credit_iban);
+    label = GTK_LABEL (gtk_builder_get_object (builder,
+                                               "provider_name_label"));
+    gtk_label_set_text (label,
+                        business);
+    label = GTK_LABEL (gtk_builder_get_object (builder,
+                                               "wire_transfer_subject_label"));
+    gtk_label_set_text (label,
+                        subject);
+    label = GTK_LABEL (gtk_builder_get_object (builder,
+                                               "amount_label"));
+    gtk_label_set_text (label,
+                        TALER_amount2s (&amount));
+  }
+
+  {
+    GtkDialog *ad;
+
+    ad = GTK_DIALOG (gtk_builder_get_object (builder,
+                                             "anastasis_gtk_c_iban_dialog"));
+    return ad;
+  }
+}
+
+
 /**
  * The user wants to solve the selected challenge. Launch the
  * dialog to allow the user to enter the solution.
@@ -2829,6 +3056,8 @@ action_challenge_solving (void)
       .ctor = &diag_code },
     { .type = gettext_noop ("email"),
       .ctor = &diag_code },
+    { .type = gettext_noop ("iban"),
+      .ctor = &diag_iban },
     { .type = NULL,
       .ctor = NULL }
   };
@@ -2888,7 +3117,8 @@ action_challenge_solving (void)
 
 
 /**
- * FIXME.
+ * The recovery process was finished. Show the recovered secret to the
+ * user.
  */
 static void
 action_recovery_finished (void)
@@ -2981,7 +3211,6 @@ action_recovery_finished (void)
   AG_hide ("anastasis_gtk_backup_complete_box");
   AG_hide ("anastasis_gtk_success_backup_label");
   AG_show ("anastasis_gtk_success_recovery_box");
-  AG_show ("anastasis_gtk_completed_image");
   AG_show ("anastasis_gtk_main_control_vbox");
   AG_hide ("anastasis_gtk_main_window_save_as_button");
   AG_show ("anastasis_gtk_restart_button");
@@ -3052,6 +3281,8 @@ AG_action_cb (void *cls,
 #endif
   if (TALER_EC_NONE != error_code)
   {
+    /* TODO: maybe also render 'detail'
+       if present in state? */
     AG_error ("Error #%d: %s\n",
               (int) error_code,
               TALER_ErrorCode_get_hint (error_code));
@@ -3060,11 +3291,14 @@ AG_action_cb (void *cls,
       GNUNET_break (0);
       return;
     }
-
   }
-  else
+  if ( (NULL != json_object_get (response,
+                                 "backup_state")) ||
+       (NULL != json_object_get (response,
+                                 "recovery_state")) )
   {
     json_decref (AG_redux_state);
+    AG_stop_long_action ();
     AG_redux_state = json_incref (response);
   }
   if ( (TALER_EC_ANASTASIS_TRUTH_UNKNOWN == error_code) ||
